@@ -30,15 +30,17 @@ import requests
 # are processed.
 #
 # This is supported in Python 3.7+
-from .exceptions import AnamlError, Reason
-from .models.summary_statistics import SummaryStatistics
-
 if typing.TYPE_CHECKING:
     import pandas
     import pyspark.sql
     import s3fs
     from google.cloud import bigquery
 
+from .exceptions import AnamlError, Reason
+from .models.summary_statistics import (
+    CategoricalSummaryStatistics, EmptySummaryStatistics, NumericalSummaryStatistics,
+    SummaryStatistics
+)
 from .models.branch import BranchName
 from .models.checks import Check, CheckCreationRequest, CheckId
 from .models.cluster import Cluster, ClusterId, ClusterName
@@ -50,7 +52,7 @@ from .models.destination import (
 from .models.destination_reference import TableDestinationReference, FolderDestinationReference
 from .models.generated_features import GeneratedFeatures
 from .models.feature import Feature
-from .models.feature_creation_request import FeatureCreationRequest
+from .models.feature_creation_request import EventFeatureCreationRequest, FeatureCreationRequest
 from .models.feature_id import FeatureId, FeatureName
 from .models.feature_run_summaries import FeatureRunSummary
 from .models.feature_set import FeatureSet, FeatureSetId, FeatureSetName
@@ -60,9 +62,19 @@ from .models.feature_template import FeatureTemplate, TemplateId, TemplateName
 from .models.feature_template_creation_request import FeatureTemplateCreationRequest
 from .models.file_format import CSV, FileFormat, Parquet, Orc
 from .models.jobs import RunStatus, FeatureStoreRunId
-from .models.merge_request import MergeRequest, MergeRequestId
+from .models.merge_request import MergeRequest, MergeRequestCreationRequest, MergeRequestId
 from .models.ref import Ref
 from .models.table import Table, TableId
+from .models.aggregate import AggregateExpression
+from .models.attribute import Attribute
+from .models.entity import Entity, EntityId, EntityName
+from .models.event_window import EventWindow
+from .models.feature_set_creation_request import FeatureSetCreationRequest
+from .models.filter_expression import FilterExpression
+from .models.label import Label
+from .models.table_preview import NullCell, TablePreview
+from .models.post_aggregate_expression import PostAggregateExpression
+from .models.select_expression import SelectExpression
 from . import version
 
 
@@ -312,7 +324,7 @@ class Anaml:
             fields updated.
         """
         r = self._put(f"/feature/{str(feature.id.value)}", json=feature.to_json())
-        _ = self._int_or_handle_errors(r)
+        _ = self._json_or_handle_errors(r)
         return self.get_feature_by_id(int(feature.id.value))
 
     def preview_feature(self, feature: Union[List[Feature], Feature]) -> List[SummaryStatistics]:
@@ -332,7 +344,7 @@ class Anaml:
             for fs in result.get('previewData', {}).get('featureStatistics', [])
         ]
 
-    def plot_feature_preview(self, feature: Feature) -> None:
+    def preview_feature_and_plot(self, feature: Union[List[Feature], Feature]) -> None:
         """Show a matplotlib plot for the preview statistics of a feature.
 
         Args:
@@ -342,8 +354,8 @@ class Anaml:
         for fs in feature_stats:
             self._build_feature_plots(fs)
         self._warn_empty_feature_stats([
-            fs["featureName"]
-            for fs in feature_stats if fs.get("adt_type") == "empty"
+            fs.featureName
+            for fs in feature_stats if isinstance(fs, EmptySummaryStatistics)
         ])
 
     def sample_feature(self, feature: Feature) -> pandas.DataFrame:
@@ -534,6 +546,32 @@ class Anaml:
         result = self._json_or_handle_errors(r)
         return FeatureSet.from_json(result)
 
+    def preview_feature_set(self, feature_set: Union[FeatureSet, FeatureSetCreationRequest]) -> pandas.DataFrame:
+        """Returns a pandas dataframe of the feature set.
+
+        Args:
+            feature_set: a FeatureSet object.
+        """
+        import pandas
+        req = {"featureSet": feature_set.to_json()}
+        r = self._post("/feature-set-preview", json=req)
+        result = self._json_or_handle_errors(r)
+        table_preview = TablePreview.from_json(result.get('previewData', {}))
+
+        # Convert to pandas
+        col_names = [c.name for c in table_preview.headers]
+        md_list = []
+        for row in table_preview.rows:
+            cell_list = []
+            for cell in row.cells:
+                if isinstance(cell, NullCell):
+                    cell_list.append(None)
+                else:
+                    cell_list.append(cell.data)
+            md_list.append(cell_list)
+
+        return pandas.DataFrame(md_list, columns=col_names)
+
     def get_run_for_feature_set(self, feature_set_id: Union[FeatureSetId, int]) -> FeatureStoreRun:
         """Get the most recent feature store run for the given feature set.
 
@@ -725,6 +763,128 @@ class Anaml:
         # Let's re-load the saved check. Just to make sure it's all up to date.
         return self.get_check(check.commit, check.id)
 
+    def get_latest_commit(self, branch: str) -> Commit:
+        """Get the latest Commit for a branch.
+
+        Args:
+            branch: The branch name.
+
+        Returns:
+            The Commit object.
+        """
+        r = self._get(f"/branch/{branch}")
+        result = self._json_or_handle_errors(r)
+        return Commit.from_json(result)
+
+    def create_merge_request(self, merge_request: MergeRequestCreationRequest) -> MergeRequestId:
+        """Create a new merge request.
+
+        Args:
+            merge_request: A new MergeRequestCreationRequest object.
+
+        Returns:
+            The new MergeRequestId object.
+        """
+        r = self._post("/merge-request", merge_request.to_json())
+        result = self._json_or_handle_errors(r)
+        return MergeRequestId.from_json(result)
+
+    # Entity Related Functions
+    def get_entity_by_id(self, entity_id: Union[EntityId, int]) -> Entity:
+        """Get a entity from the Anaml server.
+
+        Args:
+            entity_id: Unique identifier of the entity to retrieve.
+
+        Returns:
+            The requested entity, if it exists.
+        """
+        r = self._get(f"/entity/{int(entity_id)}")
+        result = self._json_or_handle_errors(r)
+        return Entity.from_json(result)
+
+    def get_entity_by_name(self, name: Union[EntityName, str]) -> Entity:
+        """Get a entity from the Anaml server.
+
+        Args:
+            name: Name of the entity to retrieve.
+
+        Returns:
+            The requested entity, if it exists.
+        """
+        r = self._get("/entity", query={'name': str(name)})
+        result = self._json_or_handle_errors(r)
+        return Entity.from_json(result)
+
+    # Builder functions
+    def build_event_feature(
+        self,
+        *,
+        name: str,
+        table: typing.Union[int, str, Table],
+        select: str,
+        aggregate: AggregateExpression,
+        window: EventWindow,
+        description: str = "",
+        labels: List[Label] = [],
+        attributes: List[Attribute] = [],
+        filter: str = None,
+        postAggregateExpr: str = None,
+        template: TemplateId = None
+    ) -> EventFeatureCreationRequest:
+        """Helper function for building event features."""
+        if isinstance(table, Table):
+            table_id = table.id
+        elif isinstance(table, str):
+            table_id = self.get_table_by_name(table).id
+        else:
+            table_id = table
+
+        if filter is not None:
+            filter = FilterExpression(filter)
+        if postAggregateExpr is not None:
+            postAggregateExpr = PostAggregateExpression(postAggregateExpr)
+
+        return EventFeatureCreationRequest(
+            name=FeatureName(name),
+            description=description,
+            attributes=attributes,
+            labels=labels,
+            table=table_id,
+            select=SelectExpression(select),
+            filter=filter,
+            aggregate=aggregate,
+            window=window,
+            postAggregateExpr=postAggregateExpr,
+            template=template
+        )
+
+    def build_feature_set(
+        self,
+        *,
+        name: str,
+        entity: typing.Union[int, str],
+        description: str = "",
+        labels: List[Label] = [],
+        attributes: List[Attribute] = [],
+        features: List[FeatureId] = []
+    ) -> FeatureSetCreationRequest:
+        """Helper function for building feature sets."""
+        if isinstance(entity, Table):
+            entity_id = entity.id
+        elif isinstance(entity, str):
+            entity_id = self.get_entity_by_name(entity).id
+        else:
+            entity_id = entity
+
+        return FeatureSetCreationRequest(
+            name=FeatureSetName(name),
+            entity=entity_id,
+            description=description,
+            labels=labels,
+            attributes=attributes,
+            features=features)
+
     #####################
     # Load feature data #
     #####################
@@ -934,14 +1094,14 @@ class Anaml:
         else:
             raise ValueError(f"Cannot load unsupported format: {format.adt_type}")
 
-    def _build_feature_plots(self, fs):
-        [self._build_numerical_plots(qdata, fs.get("featureName"))
-         for qdata in self._to_list(fs.get("quantiles"))]
-        [self._build_categorical_plots(qdata, fs.get("featureName"))
-         for qdata in self._to_list(fs.get("categoryFrequencies"))]
+    def _build_feature_plots(self, summary_stats: SummaryStatistics) -> None:
+        if isinstance(summary_stats, NumericalSummaryStatistics):
+            self._build_numerical_plots(summary_stats.quantiles, summary_stats.featureName)
+        elif isinstance(summary_stats, CategoricalSummaryStatistics):
+            self._build_categorical_plots(summary_stats.categoryFrequencies, summary_stats.featureName)
 
     @staticmethod
-    def _build_numerical_plots(qdata, title: str) -> None:
+    def _build_numerical_plots(qdata: List[float], title: str) -> None:
         import numpy
         import seaborn
         from matplotlib import pyplot
@@ -958,6 +1118,7 @@ class Anaml:
     def _build_categorical_plots(qdata, title: str) -> None:
         from matplotlib import pyplot
         import seaborn
+        import pandas
         seaborn.set_style('whitegrid')
         seaborn.catplot(x="category", y="frequency", kind="bar", data=pandas.DataFrame(qdata))
         pyplot.title(title)
@@ -969,10 +1130,6 @@ class Anaml:
     def _to_list(gotten: Optional[A]) -> List[A]:
         return [] if gotten is None else [gotten]
 
-    def _warn_empty_stats(self, features: List[str]):
-        if features:
-            self._log.warning(_NO_FEATURES.format(thefeatures=', '.join(features)))
-
     def _json_or_handle_errors(self, r: requests.Response):
         if r.ok:
             try:
